@@ -1,4 +1,4 @@
-# Copyright 2012 Dusty Phillips
+# Copyright 2012-2013 Dusty Phillips
 
 # This file is part of gitifyhg.
 
@@ -24,20 +24,46 @@ import os
 import json
 import re
 import subprocess
+import errno
 from path import path as p
+from time import strftime
+
+
+# Enable "plain" mode to make us resilient against changes to the locale, as we
+# rely on parsing certain messages produced by Mercurial. See issue #26.
+os.environ['HGPLAIN'] = '1'
+# Disable loading of the user's $HOME/.hgrc as extensions can cause weird
+# interactions and it's better to run in a known state.
+os.environ['HGRCPATH'] = ''
+
 
 from mercurial.ui import ui
 from mercurial.context import memctx, memfilectx
+from mercurial.error import Abort
 from mercurial import encoding
 from mercurial.bookmarks import listbookmarks, readcurrent, pushbookmark
 from mercurial.util import sha1
 from mercurial import hg
-from mercurial.error import RepoLookupError
+from mercurial.node import nullid
+from mercurial.node import hex as hghex  # What idiot overroad a builtin?
+from mercurial.node import short as hgshort
 
 
 DEBUG_GITIFYHG = os.environ.get("DEBUG_GITIFYHG") != None
 
 
+# hijack stdout to prevent mercurial from inadvertently talking to git.
+# interactive=off and ui.pushbuffer() don't seem to work.
+class DummyOut(object):
+    def write(self, x):
+        pass
+
+    def flush(self):
+        pass
+actual_stdout = sys.stdout
+sys.stdout = DummyOut()
+
+
 def log(msg, level="DEBUG"):
     '''The git remote operates on stdin and stdout, so all debugging information
     has to go to stderr.'''
@@ -54,7 +80,7 @@ def output(msg=''):
     if isinstance(msg, unicode):
         msg = msg.encode('utf-8')
     log("OUT: %s" % msg)
-    print(msg)
+    print >> actual_stdout, msg
 
 
 def gittz(tz):
@@ -86,29 +112,41 @@ def git_to_hg_spaces(name):
     return name.replace('___', ' ')
 
 
-AUTHOR = re.compile(r'^([^<>]+?)? ?<([^<>]*)>$')
-NAME = re.compile(r'^([^<>]+)')
+AUTHOR = re.compile(r'^([^<>]+)?(<(?:[^<>]*)>| [^ ]*@.*|[<>].*)$')
 
 
 def sanitize_author(author):
     '''Mercurial allows a more freeform user string than git, so we have to
-    massage it to be compatible. Git experts "name <email>".'''
-    name = "unknown"
-    email = "unknown"
+    massage it to be compatible. Git expects "name <email>".'''
+    name = ''
+    email = 'unknown'
     author = author.replace('"', '')
     match = AUTHOR.match(author)
     if match:
-        name = match.group(1)
-        email = match.group(2).strip()
+        if match.group(1):  # handle 'None', e.g for input "<only@email>"
+            name = match.group(1).strip()
+        email = match.group(2).translate(None, "<>").strip()
     else:
-        match = NAME.match(author)
-        if match:
-            if "@" in match.group(1):  # when they provide email without name
-                email = match.group(1).strip()
-            else:
-                name = match.group(1).strip()
+        author = author.translate(None, "<>").strip()
+        if "@" in author:
+            email = author
+        else:
+            name = author
 
-    return '%s <%s>' % (name, email)
+    if name:
+        return "%s <%s>" % (name, email)
+    else:
+        return "<%s>" % (email)
+
+
+def branch_tip(repo, branch):
+    '''HG has a lovely branch_tip method, but it requires mercurial 2.4
+    This function provides backwards compatibility. If we ever get to
+    drop older versions, we can drop this function.'''
+    if hasattr(repo, 'branchtip'):
+        return repo.branchtip(branch)
+    else:
+        return repo.branchtags()[branch]
 
 
 class HGMarks(object):
@@ -131,11 +169,13 @@ class HGMarks(object):
             self.last_mark = loaded['last-mark']
             self.marks_to_revisions = dict([(int(v), k) for k, v in
                     self.revisions_to_marks.iteritems()])
+            self.notes_mark = loaded.get('notes-mark', None)
         else:
             self.tips = {}
             self.revisions_to_marks = {}
             self.marks_to_revisions = {}
             self.last_mark = 0
+            self.notes_mark = None
 
     def store(self):
         '''Save marks to the storage file.'''
@@ -143,7 +183,8 @@ class HGMarks(object):
             json.dump({
                 'tips': self.tips,
                 'revisions_to_marks': self.revisions_to_marks,
-                'last-mark': self.last_mark},
+                'last-mark': self.last_mark,
+                'notes-mark': self.notes_mark},
             file)
 
     def mark_to_revision(self, mark):
@@ -166,6 +207,11 @@ class HGMarks(object):
     def is_marked(self, revision):
         return str(revision) in self.revisions_to_marks
 
+    def new_notes_mark(self):
+        self.last_mark += 1
+        self.notes_mark = self.last_mark
+        return self.notes_mark
+
 
 class GitRemoteParser(object):
     '''Parser for stdin that processes the git-remote protocol.'''
@@ -207,12 +253,15 @@ class GitRemoteParser(object):
         '''Read and parse an author string. Return a tuple of
         (user string, date, git_tz).'''
         self.read_line()
-        AUTHOR_RE = re.compile('^\w+ (?:(.+)? ?<.*>) (\d+) ([+-]\d+)')
+        AUTHOR_RE = re.compile(r'^(?:author|committer|tagger)(?: ([^<>]+)?)? <([^<>]*)> (\d+) ([+-]\d+)')
         match = AUTHOR_RE.match(self.line)
         if not match:
             return None
 
-        user, date, tz = match.groups()
+        user, email, date, tz = match.groups()
+        if user is None:
+            user = ''
+        user += ' <' + email + '>'
 
         date = int(date)
         tz = -((int(tz) / 100) * 3600) + ((int(tz) % 100) * 60)
@@ -235,8 +284,13 @@ class HGRemote(object):
             cmd = ['git', 'config', 'remote.%s.url' % alias, "gitifyhg::%s" % url]
             subprocess.call(cmd)
 
+        # use hash of URL as unique identifier in various places.
+        # this has the advantage over 'alias' that it stays constant
+        # when the user does a "git remote rename old new".
+        self.uuid = sha1(url.encode('utf-8')).hexdigest()
+
         gitdir = p(os.environ['GIT_DIR'].decode('utf-8'))
-        self.remotedir = gitdir.joinpath('hg', alias)
+        self.remotedir = gitdir.joinpath('hg', self.uuid)
         self.marks_git_path = self.remotedir.joinpath('marks-git')
         self.marks = HGMarks(self.remotedir.joinpath('marks-hg'))
         self.parsed_refs = {}
@@ -244,16 +298,17 @@ class HGRemote(object):
         self.branches = {}
         self.bookmarks = {}
 
-        if alias[8:] == url:  # strips off 'gitifyhg::'
-            alias = sha1(alias).hexdigest()
         self.prefix = 'refs/hg/%s' % alias
-        self.build_repo(url, alias)
+        self.alias = alias
+        self.url = url
+        self.build_repo(url)
 
-    def build_repo(self, url, alias):
+    def build_repo(self, url):
         '''Make the Mercurial repo object self.repo available. If the local
         clone does not exist, clone it, otherwise, ensure it is fetched.'''
         myui = ui()
         myui.setconfig('ui', 'interactive', 'off')
+        myui.setconfig('extensions', 'mq', '')
 
         local_path = self.remotedir.joinpath('clone')
         if not local_path.exists():
@@ -275,7 +330,7 @@ class HGRemote(object):
             if command not in ('capabilities', 'list', 'import', 'export'):
                 die('unhandled command: %s' % line)
             getattr(self, 'do_%s' % command)(parser)
-            sys.stdout.flush()
+            actual_stdout.flush()
 
         self.marks.store()
 
@@ -317,7 +372,7 @@ class HGRemote(object):
                     output()
                     return
             head = head if head != 'default' else 'master'
-            self.bookmarks[head] = node
+            #self.bookmarks[head] = node
 
         self.headnode = (head, node)
 
@@ -336,11 +391,15 @@ class HGRemote(object):
 
         # list the named branch references
         for branch in self.branches:
-            output("? refs/heads/branches/%s" % hg_to_git_spaces(branch))
+            if branch != "default":
+                output("? refs/heads/branches/%s" % hg_to_git_spaces(branch))
+            else:
+                output("? refs/heads/master")
 
         # list the bookmark references
         for bookmark in self.bookmarks:
-            output("? refs/heads/%s" % hg_to_git_spaces(bookmark))
+            if bookmark != "master":
+                output("? refs/heads/%s" % hg_to_git_spaces(bookmark))
 
         # list the tags
         for tag, node in self.repo.tagslist():
@@ -365,17 +424,20 @@ class HGImporter(object):
         self.prefix = self.hgremote.prefix
         self.repo = self.hgremote.repo
         self.parser = parser
+        self.notes_committed = 0
 
     def process(self):
         output("feature done")
         if self.hgremote.marks_git_path.exists():
             output("feature import-marks=%s" % self.hgremote.marks_git_path)
         output("feature export-marks=%s" % self.hgremote.marks_git_path)
-        sys.stdout.flush()
+        output("feature notes")
+        actual_stdout.flush()
 
         tmp = encoding.encoding
         encoding.encoding = 'utf-8'
 
+        self.commit_count = 0
         while self.parser.line.startswith('import'):
             ref = self.parser.line.split()[1]
 
@@ -384,6 +446,8 @@ class HGImporter(object):
                     self.hgremote.headnode[0],
                     'bookmarks',
                     self.hgremote.headnode[1])
+            elif ref == 'refs/heads/master':
+                self.do_branch('default')
             elif ref.startswith('refs/heads/branches/'):
                 self.do_branch(ref[len('refs/heads/branches/'):])
             elif ref.startswith('refs/heads/'):
@@ -395,6 +459,8 @@ class HGImporter(object):
                 tag = ref[len('refs/tags/'):]
                 self.process_ref(tag, 'tags', self.repo[git_to_hg_spaces(tag)])
 
+            self.process_notes()
+
             self.parser.read_line()
 
         encoding.encoding = tmp
@@ -406,7 +472,7 @@ class HGImporter(object):
             heads = self.hgremote.branches[branch]
             if len(heads) > 1:
                 log("Branch '%s' has more than one head, consider merging" % branch, "WARNING")
-                tip = self.repo.branchtip(branch)
+                tip = branch_tip(self.repo, branch)
             else:
                 tip = heads[0]
 
@@ -416,20 +482,47 @@ class HGImporter(object):
         head = self.repo[tip]
         self.process_ref(hg_to_git_spaces(branch), 'branches', head)
 
+    def process_notes(self):
+        last_notes_mark = self.marks.notes_mark if self.marks.notes_mark is not None else 0
+        mark_to_hgsha1 = [(mark, self.repo[rev].hex()) for rev, mark in
+                          self.marks.revisions_to_marks.iteritems() if mark > last_notes_mark]
+        if not mark_to_hgsha1 or self.commit_count < 1:
+            return
+        output("commit refs/notes/hg-%s" % (self.hgremote.uuid))
+        output("mark :%d" % (self.marks.new_notes_mark()))
+        output("committer <gitifyhg-note> %s" % (strftime('%s %z')))
+        message = u"hg from %s (%s)\n" % (self.prefix, self.hgremote.url)
+        message = message.encode("utf-8")
+        output("data %d" % (len(message)))
+        output(message)
+        if last_notes_mark > 0:
+            output("from :%d" % (last_notes_mark))
+        for mark, hgsha1 in mark_to_hgsha1:
+            output("N inline :%d" % (mark))
+            output("data 40")
+            output(hgsha1)
+        output()
+
     def process_ref(self, name, kind, head):
 
+        # FIXME: I really need a better variable name here.
         kind_name = "%s/%s" % (kind, name)
+        if kind_name == "branches/default":
+            # I have no idea where 'bookmarks' comes from in this case.
+            # I don't think there is meant to be many bookmarks/master ref,
+            # but this is what I had to do to make tests pass when special
+            # casing the master/default dichotomy. Something is still fishy
+            # here, but it's less fishy than it was. See issue #34.
+            kind_name = "bookmarks/master"
         tip = self.marks.tips.get(kind_name, 0)
 
-        if tip and tip == head.rev():
-            return  # shortcut for no changes
-
         revs = xrange(tip, head.rev() + 1)
         count = 0
 
-        revs = [rev for rev in revs if not self.marks.is_marked(rev)]
-
         for rev in revs:
+            if self.marks.is_marked(rev):
+                continue
+
             (manifest, user, (time, tz), files, description, extra
                 ) = self.repo.changelog.read(self.repo[rev].node())
 
@@ -485,10 +578,11 @@ class HGImporter(object):
 
         # make sure the ref is updated
         output("reset %s/%s" % (self.prefix, kind_name))
-        output("from :%u" % self.marks.revision_to_mark(rev))
+        output("from :%u" % self.marks.revision_to_mark(head.rev()))
         output()
 
-        self.marks.tips[kind_name] = rev
+        self.marks.tips[kind_name] = head.rev()
+        self.commit_count += count
 
     def get_filechanges(self, context, parent):
         modified = set()
@@ -521,11 +615,15 @@ class GitExporter(object):
         self.hgremote = hgremote
         self.marks = self.hgremote.marks
         self.parsed_refs = self.hgremote.parsed_refs
+        self.parsed_tags = {}  # refs to tuple of (message, author)
         self.blob_marks = self.hgremote.blob_marks
         self.repo = self.hgremote.repo
         self.parser = parser
+        self.processed_marks = set()
+        self.processed_nodes = []
 
     def process(self):
+        self.marks.store()  # checkpoint
         new_branch = False
         self.parser.read_line()
         for line in self.parser.read_block('done'):
@@ -546,19 +644,49 @@ class GitExporter(object):
                 if not pushbookmark(self.repo, bookmark, old, node):
                     continue
             elif ref.startswith('refs/tags/'):
-                tag = ref[len('refs/tags/'):]
-                self.repo.tag([tag], node, None, True, None, {})
-                # FIXME: the new tag needs to be committed in such a way that
-                # the commit doesn't interfere with any other commits being
-                # exported.
+                self.write_tag(ref)
             else:
                 # transport-helper/fast-export bugs
+                log("Fast-export unexpected ref: %s" % ref, "WARNING")
                 continue
-            output("ok %s" % ref)
 
+        success = False
+        try:
+            self.repo.push(self.hgremote.peer, force=False, newbranch=new_branch)
+            self.marks.store()
+            success = True
+        except Abort as e:
+            # mercurial.error.Abort: push creates new remote head f14531ca4e2d!
+            if e.message.startswith("push creates new remote head"):
+                self.marks.load()  # restore from checkpoint
+                # strip revs, implementation finds min revision from list
+                if self.processed_nodes:
+                    self.repo.mq.strip(self.repo, self.processed_nodes)
+            else:
+                die("unknown hg exception: %s" % e)
+        # TODO: handle network/other errors?
+
+        for ref in self.parsed_refs:
+            if success:
+                output("ok %s" % ref)
+            else:
+                output("error %s non-fast forward" % ref)  # TODO: other errors as well
         output()
 
-        self.repo.push(self.hgremote.peer, force=False, newbranch=new_branch)
+        if not success:
+            # wait until fast-export finishes to muck with the marks file
+            self.remove_processed_git_marks()
+
+    def remove_processed_git_marks(self):
+        with self.hgremote.marks_git_path.open() as fread:
+            with self.hgremote.marks_git_path.open('r+') as fwrite:
+                for line in fread:
+                    if not line.startswith(':'):
+                        die("invalid line in marks-git: " + line)
+                    mark = line[1:].split()[0]
+                    if mark not in self.processed_marks:
+                        fwrite.write(line)
+                fwrite.truncate()
 
     def do_blob(self):
         mark = self.parser.read_mark()
@@ -585,33 +713,6 @@ class GitExporter(object):
         from_mark = merge_mark = None
 
         ref = self.parser.line.split()[1]
-        # FIXME: This needs to be in it's own method or function somehow.
-        #        The importer uses similar logic.
-        if ref.startswith('refs/heads/branches/'):
-            branch_name = ref[len('refs/heads/branches/'):]
-            try:
-                tip = self.repo.branchtip(git_to_hg_spaces(branch_name))
-                git_marked_tip = self.marks.tips['branches/%s' % branch_name]
-            except RepoLookupError:
-                # setting these to 0 isn't honest, but it (currently)
-                # only has to pass the if git_marked_tip < tip: below
-                tip = git_marked_tip = 0
-        elif ref == 'refs/heads/master':
-            try:
-                tip = self.repo.branchtip('default')
-                git_marked_tip = self.marks.tips['bookmarks/master']
-            except RepoLookupError:
-                tip = git_marked_tip = 0
-        elif ref.startswith('refs/heads/'):
-            bookmark = ref[len('refs/heads/'):]
-            tip = listbookmarks(self.repo)[git_to_hg_spaces(bookmark)]
-            git_marked_tip = self.marks.tips['bookmarks/%s' % bookmark]
-
-        tip = self.repo[tip].rev()
-        log("%r %r" % (git_marked_tip, tip))
-        if git_marked_tip < tip:
-            output("error %s non-fast forward\n" % ref)
-            sys.exit(1)
 
         commit_mark = self.parser.read_mark()
         author = self.parser.read_author()
@@ -662,7 +763,8 @@ class GitExporter(object):
                     files[file] = {'ctx': self.repo[parent_from][file]}
 
         if ref.startswith('refs/heads/branches/'):
-            extra['branch'] = git_to_hg_spaces(branch_name)
+            extra['branch'] = git_to_hg_spaces(
+                ref[len('refs/heads/branches/'):])
 
         def get_filectx(repo, memctx, file):
             filespec = files[file]
@@ -688,16 +790,66 @@ class GitExporter(object):
 
         self.parsed_refs[ref] = node
         self.marks.new_mark(rev, commit_mark)
+        self.processed_marks.add(str(commit_mark))
+        self.processed_nodes.append(node)
 
     def do_tag(self):
-        name = self.parser.line().split()[1]
-        from_mark = self.parser.read_mark()
+        name = self.parser.line.split()[1]
+        self.parser.read_mark()
         tagger = self.parser.read_author()
-        data = self.parser.read_data()
+        message = self.parser.read_data()
+        self.parser.read_line()
+        self.parsed_tags[git_to_hg_spaces(name)] = tagger, message
 
     def do_feature(self):
         pass  # Ignore
 
+    def write_tag(self, ref):
+        node = self.parsed_refs[ref]
+        tag = git_to_hg_spaces(ref[len('refs/tags/'):])
+        # Calling self.repo.tag() doesn't append the tag to the correct
+        # commit. So I copied some of localrepo._tag into here.
+        # But that method, like much of mercurial's code, is ugly.
+        # So I then rewrote it.
+
+        try:
+            fp = self.repo.wfile('.hgtags', 'rb+')
+        except IOError as e:
+            if e.errno != errno.ENOENT:
+                raise
+            fp = self.repo.wfile('.hgtags', 'ab')
+            prevtags = ''
+        else:
+            prevtags = fp.read()
+
+        fp.seek(0, 2)
+        if prevtags and prevtags[-1] != '\n':
+            fp.write('\n')
+        encoded_tag = encoding.fromlocal(tag)
+        fp.write('%s %s\n' % (hghex(node), encoded_tag))
+        fp.close()
+
+        def get_filectx(repo, memctx, file):
+            return memfilectx(file, repo.wfile('.hgtags', 'rb').read())
+        branch_tag = self.repo[node].branch()
+        if tag in self.parsed_tags:
+            author, message = self.parsed_tags[tag]
+            user, date, tz = author
+            date_tz = (date, tz)
+        else:
+            message = "Added tag %s for changeset %s" % (tag, hgshort(node))
+            user = None
+            date_tz = None
+        ctx = memctx(self.repo,
+            (branch_tip(self.repo, branch_tag), self.NULL_PARENT),
+            message,
+            ['.hgtags'], get_filectx, user, date_tz, {'branch': branch_tag})
+
+        tmp = encoding.encoding
+        encoding.encoding = 'utf-8'
+        node = self.repo.commitctx(ctx)
+        encoding.encoding = tmp
+
 
 def main():
     '''Main entry point for the git-remote-gitifyhg command. Parses sys.argv
