@@ -21,7 +21,8 @@ import sys
 import pytest
 import sh
 from .helpers import (make_hg_commit, make_git_commit, clone_repo,
-    assert_hg_count)
+    assert_hg_count, assert_git_count, assert_hg_messages,
+    assert_hg_author, assert_git_author)
 
 
 def test_simple_push_from_master(hg_repo, git_dir):
@@ -36,6 +37,14 @@ def test_simple_push_from_master(hg_repo, git_dir):
         assert file.read() == "a\nb"
 
 
+def test_simple_push_updates_remote(hg_repo, git_dir):
+    clone_repo(git_dir, hg_repo)
+    make_git_commit("b")
+    sh.git.push()
+    sh.git.fetch()
+    assert_git_count(2, ref='origin')
+
+
 def test_empty_repo(tmpdir):
     tmpdir = p(tmpdir.strpath).abspath()
     hg_base = tmpdir.joinpath('hg_base')
@@ -68,7 +77,6 @@ def test_push_conflict_default(git_dir, hg_repo):
     assert sh.git.push(_ok_code=1).stderr.find("master -> master (non-fast-forward)") > 0
 
 
-@pytest.mark.xfail
 def test_push_conflict_default_double(git_dir, hg_repo):
     git_repo = clone_repo(git_dir, hg_repo)
     sh.cd(hg_repo)
@@ -79,6 +87,17 @@ def test_push_conflict_default_double(git_dir, hg_repo):
     assert sh.git.push(_ok_code=1).stderr.find("master -> master (non-fast-forward)") > 0
 
 
+def test_push_conflict_default_double_non_english(git_dir, hg_repo, monkeypatch):
+    monkeypatch.setenv('LANG', 'de_DE')
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(hg_repo)
+    make_hg_commit("b")
+    sh.cd(git_repo)
+    make_git_commit("c")
+    assert sh.git.push(_ok_code=1).stderr.find("master -> master (non-fast-forward)") > 0
+    assert sh.git.push(_ok_code=1).stderr.find("master -> master (non-fast-forward)") > 0
+
+
 def test_push_to_named(git_dir, hg_repo):
     sh.cd(hg_repo)
     sh.hg.branch("branch_one")
@@ -228,6 +247,81 @@ def test_push_with_multiple_bookmarks(git_dir, hg_repo):
         assert file.read() == "a\nbd"
 
 
+@pytest.mark.xfail
+def test_fetch_after_bad_push_updates_origin_master(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(hg_repo)
+    make_hg_commit("b")
+    sh.cd(git_repo)
+    make_git_commit("c", "c")
+    try:
+        sh.git.push()
+    except sh.ErrorReturnCode as ret:
+        print("Output from failed push:")
+        print('\t' + '\n\t'.join(ret[0].splitlines()))
+    sh.git.fetch()
+    assert_git_count(2, ref="origin/branches/default")
+    assert_git_count(2, ref="origin/master")
+    sh.git.pull('--rebase')
+    assert_git_count(3)         # as with test_pull_auto_merge
+    sh.git.push()
+    sh.cd(hg_repo)
+    assert_hg_count(3)
+
+
+def test_push_after_rebase(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(hg_repo)
+    make_hg_commit("b")
+    sh.cd(git_repo)
+    make_git_commit("c", "c")
+    sh.git.pull('--rebase')
+    assert_git_count(3)         # as with test_pull_auto_merge
+    sh.git.push()
+    sh.cd(hg_repo)
+    assert_hg_count(3)
+
+
+def test_push_email(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(git_repo)
+    make_git_commit("b")
+    sh.git.push()
+    sh.cd(hg_repo)
+    sh.hg.update()
+    assert_hg_author()
+    hg_user = "Hg Author <hg@author.test>"
+    make_hg_commit("c", user=hg_user)
+    assert_hg_author(hg_user)
+    sh.cd(git_repo)
+    sh.git.pull()
+    assert_git_author(ref="HEAD^")
+    assert_git_author(author=hg_user)
+
+
+def test_push_after_merge(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(hg_repo)
+    make_hg_commit("b")
+    sh.cd(git_repo)
+    make_git_commit("c", "c")
+    sh.git.pull()               # automerge
+    assert_git_count(4)         # as with test_pull_auto_merge
+    sh.git.push()
+    sh.cd(hg_repo)
+    assert_hg_count(4)
+
+
+def test_push_two_commits(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(git_repo)
+    make_git_commit("b")
+    make_git_commit("c")
+    sh.git.push()
+    sh.cd(hg_repo)
+    assert_hg_count(3)
+
+
 @pytest.mark.xfail
 def test_push_new_bookmark(git_dir, hg_repo):
     git_repo = clone_repo(git_dir, hg_repo)
@@ -247,29 +341,66 @@ def test_push_new_bookmark(git_dir, hg_repo):
     # to create a named branch. This needs to be fixed.
 
 
-@pytest.mark.xfail
 def test_push_tag(git_dir, hg_repo):
     git_repo = clone_repo(git_dir, hg_repo)
     sh.cd(git_repo)
     sh.git.tag("this_is_a_tag")
-    sh.git.push(tags=True)
+    sh.git.push(tags=True, _err=sys.stderr)
 
     sh.cd(hg_repo)
     assert "this_is_a_tag" in sh.hg.tags().stdout
+    assert_hg_count(2)
 
-    # TODO: this currently fails because the hg repository needs a new commit
-    # after hg tag is called.
+
+def test_push_tag_with_subsequent_commits(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(git_repo)
+    sh.git.tag("this_is_a_tag")
+    make_git_commit("b")
+    sh.git.push("origin", "HEAD", tags=True, _err=sys.stderr)
+
+    sh.cd(hg_repo)
+    assert "this_is_a_tag" in sh.hg.tags().stdout
+    assert_hg_count(3)
+    hgsha = sh.hg.id(id=True).stdout.strip()
+    assert_hg_messages(['Added tag this_is_a_tag for changeset %s' % hgsha,
+        'b', 'a'])
+
+
+def test_push_messaged_tag(git_dir, hg_repo):
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(git_repo)
+    sh.git.tag("this_is_a_tag", message="I tagged this with a message and user")
+    sh.git.push(tags=True, _err=sys.stderr)
+
+    sh.cd(hg_repo)
+    assert "this_is_a_tag" in sh.hg.tags().stdout
+    assert_hg_count(2)
+    assert_hg_messages(['I tagged this with a message and user', 'a'])
+
+
+def test_push_tag_different_branch(git_dir, hg_repo):
+    sh.cd(hg_repo)
+    sh.hg.branch("branch_one")
+    make_hg_commit("b")
+    git_repo = clone_repo(git_dir, hg_repo)
+    sh.cd(git_repo)
+    sh.git.checkout("origin/branches/branch_one", track=True)
+    sh.git.tag("this_is_a_tag")
+    sh.git.push(tags=True, _err=sys.stderr)
+
+    sh.cd(hg_repo)
+    assert "this_is_a_tag" in sh.hg.tags().stdout
+    assert_hg_count(3)
+    sh.hg.update('tip')  # tip is the commit that added the tag
+    assert sh.hg.branch().stdout.strip() == "branch_one"
 
 
-@pytest.mark.xfail
 def test_push_tag_with_spaces(git_dir, hg_repo):
     git_repo = clone_repo(git_dir, hg_repo)
     sh.cd(git_repo)
-    sh.git.tag("this is a tag")
+    sh.git.tag("this___is___a___tag")
     sh.git.push(tags=True)
 
     sh.cd(hg_repo)
-    assert "this___is___a___tag" in sh.hg.tags().stdout
-
-    # TODO: this currently fails because the hg repository needs a new commit
-    # after hg tag is called.
+    assert "this is a tag" in sh.hg.tags().stdout
